@@ -272,6 +272,50 @@ protected:
   }
 };
 
+class ReducescatterOp : public HorovodOp {
+public:
+  explicit ReducescatterOp(HorovodGlobalState* global_state);
+
+  virtual ~ReducescatterOp() = default;
+
+  Status Execute(std::vector<TensorTableEntry>& entries,
+                 const Response& response) override = 0;
+
+  virtual bool Enabled(const ParameterManager& param_manager,
+                       const std::vector<TensorTableEntry>& entries,
+                       const Response& response) const = 0;
+
+protected:
+  virtual TensorShape ComputeOutputShapeForRank(const TensorShape& tensor_shape,
+                                                int rank,
+                                                int global_size) const;
+
+  virtual std::vector<std::vector<TensorShape>>
+  ComputeOutputShapes(const std::vector<TensorTableEntry>& entries,
+                      int global_size) const;
+
+  virtual std::vector<int> ComputeReceiveCounts(
+      const std::vector<std::vector<TensorShape>>& output_shapes) const;
+
+  virtual Status AllocateOutput(std::vector<TensorTableEntry>& entries,
+                                const std::vector<TensorShape>& output_shapes);
+
+  virtual void MemcpyInFusionBuffer(
+      const std::vector<TensorTableEntry>& entries,
+      const std::vector<std::vector<TensorShape>>& output_shapes,
+      std::size_t element_size, void*& buffer_data);
+
+  virtual void MemcpyOutFusionBuffer(const void* buffer_data,
+                                     std::vector<TensorTableEntry>& entries);
+
+  virtual void MemcpyEntryInFusionBuffer(const TensorTableEntry& e,
+                                         size_t entry_offset, size_t entry_size,
+                                         void* buffer_data_at_offset);
+
+  virtual void MemcpyEntryOutFusionBuffer(const void* buffer_data_at_offset,
+                                          TensorTableEntry& e);
+};
+
 class JoinOp : public HorovodOp {
 public:
   explicit JoinOp(HorovodGlobalState* global_state);
