@@ -20,11 +20,17 @@
 namespace horovod {
 namespace common {
 
-MPIAllreduce::MPIAllreduce(MPIContext* mpi_context, HorovodGlobalState* global_state)
-    : AllreduceOp(global_state), mpi_context_(mpi_context) {}
+MPIAllreduce::MPIAllreduce(HorovodGlobalState* global_state)
+    : AllreduceOp(global_state) {}
 
 Status MPIAllreduce::Execute(std::vector<TensorTableEntry>& entries, const Response& response) {
+  assert(!entries.empty());
+  WaitForData(entries);
+
   auto& first_entry = entries[0];
+  auto& process_set =
+      global_state_->process_set_table.Get(first_entry.process_set_id);
+  const auto& mpi_context = process_set.mpi_context;
 
   const void* fused_input_data;
   void* buffer_data;
@@ -53,11 +59,11 @@ Status MPIAllreduce::Execute(std::vector<TensorTableEntry>& entries, const Respo
   timeline.ActivityStartAll(entries, MPI_ALLREDUCE);
   const void* sendbuf = entries.size() > 1 || fused_input_data == buffer_data
                         ? MPI_IN_PLACE : fused_input_data;
-  int op = MPI_Allreduce(sendbuf, buffer_data,
-                         (int) num_elements,
-                         mpi_context_->GetMPIDataType(first_entry.tensor),
-                         mpi_context_->GetMPISumOp(first_entry.tensor->dtype()),
-                         mpi_context_->GetMPICommunicator(Communicator::GLOBAL));
+  int op =
+      MPI_Allreduce(sendbuf, buffer_data, (int)num_elements,
+                    mpi_context.GetMPIDataType(first_entry.tensor),
+                    mpi_context.GetMPISumOp(first_entry.tensor->dtype()),
+                    mpi_context.GetMPICommunicator(Communicator::GLOBAL));
   if (op != MPI_SUCCESS) {
     throw std::runtime_error("MPI_Allreduce failed, see MPI output for details.");
   }
@@ -84,8 +90,8 @@ bool MPIAllreduce::Enabled(const ParameterManager& param_manager,
   return true;
 }
 
-MPIAllgather::MPIAllgather(MPIContext* mpi_context, HorovodGlobalState* global_state)
-    : AllgatherOp(global_state), mpi_context_(mpi_context) {}
+MPIAllgather::MPIAllgather(HorovodGlobalState* global_state)
+    : AllgatherOp(global_state) {}
 
 bool MPIAllgather::Enabled(const ParameterManager& param_manager,
                            const std::vector<TensorTableEntry>& entries,
@@ -94,6 +100,14 @@ bool MPIAllgather::Enabled(const ParameterManager& param_manager,
 }
 
 Status MPIAllgather::Execute(std::vector<TensorTableEntry>& entries, const Response& response) {
+  assert(!entries.empty());
+  WaitForData(entries);
+
+  auto& first_entry = entries[0];
+  auto& process_set =
+      global_state_->process_set_table.Get(first_entry.process_set_id);
+  const auto& mpi_context = process_set.mpi_context;
+
   auto& timeline = global_state_->timeline;
 
   // Sizes of subcomponents of each entry from all ranks
@@ -103,7 +117,7 @@ Status MPIAllgather::Execute(std::vector<TensorTableEntry>& entries, const Respo
   // allgatherv
   auto** entry_component_offsets = new int64_t* [entries.size()];
 
-  int global_size = global_state_->controller->GetSize();
+  int global_size = process_set.controller->GetSize();
   auto* recvcounts = new int[global_size]();
   auto* displcmnts = new int[global_size]();
 
@@ -112,8 +126,6 @@ Status MPIAllgather::Execute(std::vector<TensorTableEntry>& entries, const Respo
     entry_component_offsets[ec] = new int64_t[global_size]();
   }
 
-  auto& first_entry = entries[0];
-
   timeline.ActivityStartAll(entries, ALLOCATE_OUTPUT);
   Status status = AllocateOutput(entries, response, entry_component_sizes, recvcounts);
   if (!status.ok()) {
@@ -121,7 +133,7 @@ Status MPIAllgather::Execute(std::vector<TensorTableEntry>& entries, const Respo
     for (size_t ec = 0; ec < entries.size(); ++ec) {
       delete[] entry_component_sizes[ec];
       delete[] entry_component_offsets[ec];
-    }   
+    }
     delete[] entry_component_sizes;
     delete[] entry_component_offsets;
     delete[] recvcounts;
@@ -130,10 +142,10 @@ Status MPIAllgather::Execute(std::vector<TensorTableEntry>& entries, const Respo
   }
   timeline.ActivityEndAll(entries);
 
-  SetDisplacements(recvcounts, displcmnts);
+  SetDisplacements(recvcounts, displcmnts, global_size);
   SetEntryComponentOffsets(entries, entry_component_sizes, recvcounts, entry_component_offsets);
 
-  int element_size = mpi_context_->GetMPITypeSize(first_entry.tensor->dtype());
+  int element_size = mpi_context.GetMPITypeSize(first_entry.tensor->dtype());
 
   const void* sendbuf = nullptr;
   void* buffer_data;
@@ -149,7 +161,7 @@ Status MPIAllgather::Execute(std::vector<TensorTableEntry>& entries, const Respo
   }
 
   global_state_->timeline.ActivityStartAll(entries, MPI_ALLGATHER);
-  auto dtype = mpi_context_->GetMPIDataType(first_entry.tensor->dtype());
+  auto dtype = mpi_context.GetMPIDataType(first_entry.tensor->dtype());
   int op = MPI_Allgatherv(sendbuf != nullptr ? sendbuf : MPI_IN_PLACE,
                           (int) total_num_elements,
                           dtype,
@@ -157,7 +169,7 @@ Status MPIAllgather::Execute(std::vector<TensorTableEntry>& entries, const Respo
                           recvcounts,
                           displcmnts,
                           dtype,
-                          mpi_context_->GetMPICommunicator(Communicator::GLOBAL));
+                          mpi_context.GetMPICommunicator(Communicator::GLOBAL));
   if (op != MPI_SUCCESS) {
     throw std::runtime_error("MPI_Allgatherv failed, see MPI output for details.");
   }
@@ -183,11 +195,19 @@ Status MPIAllgather::Execute(std::vector<TensorTableEntry>& entries, const Respo
   return Status::OK();
 }
 
-MPIHierarchicalAllgather::MPIHierarchicalAllgather(MPIContext* mpi_context,
-                                                   HorovodGlobalState* global_state)
-    : MPIAllgather(mpi_context, global_state) {}
+MPIHierarchicalAllgather::MPIHierarchicalAllgather(
+    HorovodGlobalState* global_state)
+    : MPIAllgather(global_state) {}
 
 Status MPIHierarchicalAllgather::Execute(std::vector<TensorTableEntry>& entries, const Response& response) {
+  assert(!entries.empty());
+  WaitForData(entries);
+
+  auto& first_entry = entries[0];
+  auto& process_set =
+      global_state_->process_set_table.Get(first_entry.process_set_id);
+  auto& mpi_context = process_set.mpi_context;
+
   auto& timeline = global_state_->timeline;
 
   // Sizes of subcomponents of each entry from all ranks
@@ -197,7 +217,7 @@ Status MPIHierarchicalAllgather::Execute(std::vector<TensorTableEntry>& entries,
   // allgatherv
   auto** entry_component_offsets = new int64_t* [entries.size()];
 
-  int global_size = global_state_->controller->GetSize();
+  int global_size = process_set.controller->GetSize();
   auto* recvcounts = new int[global_size]();
   auto* displcmnts = new int[global_size]();
 
@@ -206,8 +226,6 @@ Status MPIHierarchicalAllgather::Execute(std::vector<TensorTableEntry>& entries,
     entry_component_offsets[ec] = new int64_t[global_size]();
   }
 
-  auto& first_entry = entries[0];
-
   timeline.ActivityStartAll(entries, ALLOCATE_OUTPUT);
   Status status = AllocateOutput(entries, response, entry_component_sizes, recvcounts);
   if (!status.ok()) {
@@ -215,7 +233,7 @@ Status MPIHierarchicalAllgather::Execute(std::vector<TensorTableEntry>& entries,
     for (size_t ec = 0; ec < entries.size(); ++ec) {
       delete[] entry_component_sizes[ec];
       delete[] entry_component_offsets[ec];
-    }   
+    }
     delete[] entry_component_sizes;
     delete[] entry_component_offsets;
     delete[] recvcounts;
@@ -224,111 +242,112 @@ Status MPIHierarchicalAllgather::Execute(std::vector<TensorTableEntry>& entries,
   }
   timeline.ActivityEndAll(entries);
 
-  SetDisplacements(recvcounts, displcmnts);
+  SetDisplacements(recvcounts, displcmnts, global_size);
   SetEntryComponentOffsets(entries, entry_component_sizes, recvcounts, entry_component_offsets);
 
-  int element_size = mpi_context_->GetMPITypeSize(first_entry.tensor->dtype());
+  int element_size = mpi_context.GetMPITypeSize(first_entry.tensor->dtype());
 
   int64_t total_size = displcmnts[global_size - 1] +
                        recvcounts[global_size - 1];
 
   // If shared buffer is not initialized or is not large enough, reallocate
   int64_t total_size_in_bytes = total_size * element_size;
-  if (global_state_->shared_buffer == nullptr || global_state_->shared_buffer_size < total_size_in_bytes) {
-    if (global_state_->shared_buffer != nullptr) {
-      MPI_Win_fence(0, mpi_context_->window);
-      MPI_Win_free(&mpi_context_->window);
-      global_state_->shared_buffer = nullptr;
+  if (process_set.shared_buffer == nullptr ||
+      process_set.shared_buffer_size < total_size_in_bytes) {
+    if (process_set.shared_buffer != nullptr) {
+      MPI_Win_fence(0, mpi_context.window);
+      MPI_Win_free(&mpi_context.window);
+      process_set.shared_buffer = nullptr;
     }
 
     // Allocate shared memory, give each rank their respective pointer
     timeline.ActivityStartAll(entries, ALLOCATE_SHARED_BUFFER);
-    int64_t window_size = global_state_->controller->GetLocalRank() == 0 ? total_size_in_bytes : 0;
+    int64_t window_size = process_set.controller->GetLocalRank() == 0 ? total_size_in_bytes : 0;
     MPI_Win_allocate_shared(window_size,
                             element_size,
                             MPI_INFO_NULL,
-                            mpi_context_->GetMPICommunicator(Communicator::LOCAL),
-                            &global_state_->shared_buffer,
-                            &mpi_context_->window);
-    if (global_state_->controller->GetLocalRank() != 0) {
+                            mpi_context.GetMPICommunicator(Communicator::LOCAL),
+                            &process_set.shared_buffer,
+                            &mpi_context.window);
+    if (process_set.controller->GetLocalRank() != 0) {
       int disp_unit;
       MPI_Aint winsize;
-      MPI_Win_shared_query(mpi_context_->window,
+      MPI_Win_shared_query(mpi_context.window,
                            0,
                            &winsize,
                            &disp_unit,
-                           &global_state_->shared_buffer);
+                           &process_set.shared_buffer);
     }
-    global_state_->shared_buffer_size = total_size_in_bytes;
+    process_set.shared_buffer_size = total_size_in_bytes;
     timeline.ActivityEndAll(entries);
   }
 
   // Compute cross-node allgather displacements and recvcounts for
   // homogeneous/parallelized case
-  int cross_size = global_state_->controller->GetCrossSize();
-  int local_size = global_state_->controller->GetLocalSize();
-  int local_rank = global_state_->controller->GetLocalRank();
+  int cross_size = process_set.controller->GetCrossSize();
+  int local_size = process_set.controller->GetLocalSize();
+  int local_rank = process_set.controller->GetLocalRank();
   auto* cross_recvcounts = new int[cross_size]();
   auto* cross_displcmnts = new int[cross_size]();
 
-  if (global_state_->controller->IsHomogeneous()) {
-    for (int i = 0; i < global_state_->controller->GetCrossSize(); ++i) {
+  if (process_set.controller->IsHomogeneous()) {
+    for (int i = 0; i < process_set.controller->GetCrossSize(); ++i) {
       cross_recvcounts[i] = recvcounts[local_size * i + local_rank];
       cross_displcmnts[i] = displcmnts[local_size * i + local_rank];
     }
-  } else if (global_state_->controller->GetLocalRank() == 0) {
+  } else if (process_set.controller->GetLocalRank() == 0) {
     // In this case local rank 0 will allgather with all local data
     int offset = 0;
     for (int i = 0; i < cross_size; ++i) {
-      for (int j = offset; j < offset + global_state_->controller->GetLocalSizeAtCrossRank(i);
+      for (int j = offset; j < offset + process_set.controller->GetLocalSizeAtCrossRank(i);
            ++j) {
         cross_recvcounts[i] += recvcounts[j];
       }
       cross_displcmnts[i] = displcmnts[offset];
-      offset += global_state_->controller->GetLocalSizeAtCrossRank(i);
+      offset += process_set.controller->GetLocalSizeAtCrossRank(i);
     }
   }
 
   timeline.ActivityStartAll(entries, MEMCPY_IN_SHARED_BUFFER);
 
-  int rank = global_state_->controller->GetRank();
+  int rank = process_set.controller->GetRank();
   for (size_t ec = 0; ec < entries.size(); ++ec) {
     auto& e = entries[ec];
     void* shared_buffer_at_offset =
-        (uint8_t*) global_state_->shared_buffer +
+        (uint8_t*)process_set.shared_buffer +
         entry_component_offsets[ec][rank] * element_size;
 
     // CPU copy to shared buffer
     memcpy(shared_buffer_at_offset, e.tensor->data(),
            (size_t) (entry_component_sizes[ec][rank] * element_size));
   }
-  Barrier();
+  Barrier(process_set);
   timeline.ActivityEndAll(entries);
 
   // Perform the cross-node allgather. If the cluster is homogeneous all
   // local ranks participate, otherwise local rank 0 handles all data
   global_state_->timeline.ActivityStartAll(entries, MPI_CROSS_ALLGATHER);
-  if (global_state_->controller->IsHomogeneous() || global_state_->controller->GetLocalRank() == 0) {
+  if (process_set.controller->IsHomogeneous() || process_set.controller->GetLocalRank() == 0) {
     int op = MPI_Allgatherv(MPI_IN_PLACE,
                             0,
                             MPI_DATATYPE_NULL,
-                            global_state_->shared_buffer,
+                            process_set.shared_buffer,
                             cross_recvcounts,
                             cross_displcmnts,
-                            mpi_context_->GetMPIDataType(first_entry.tensor->dtype()),
-                            mpi_context_->GetMPICommunicator(Communicator::CROSS));
+                            mpi_context.GetMPIDataType(first_entry.tensor->dtype()),
+                            mpi_context.GetMPICommunicator(Communicator::CROSS));
     if (op != MPI_SUCCESS) {
       throw std::runtime_error("MPI_Allgatherv failed, see MPI output for details.");
     }
   }
-  Barrier();
+  Barrier(process_set);
   global_state_->timeline.ActivityEndAll(entries);
 
   // Copy memory out of the fusion buffer.
   timeline.ActivityStartAll(entries, MEMCPY_OUT_FUSION_BUFFER);
   MemcpyOutFusionBuffer(entry_component_offsets, entry_component_sizes,
-                        global_state_->shared_buffer, element_size, entries);
-  Barrier();
+                        process_set.shared_buffer, element_size, entries);
+  Barrier(process_set);
   timeline.ActivityEndAll(entries);
 
   // Free the buffers
@@ -344,23 +363,29 @@ bool MPIHierarchicalAllgather::Enabled(const ParameterManager& param_manager,
   return param_manager.HierarchicalAllgather();
 }
 
-void MPIHierarchicalAllgather::Barrier() {
-  int op = MPI_Barrier(mpi_context_->GetMPICommunicator(Communicator::GLOBAL));
+void MPIHierarchicalAllgather::Barrier(const ProcessSet& process_set) {
+  int op =
+      MPI_Barrier(process_set.mpi_context.GetMPICommunicator(Communicator::GLOBAL));
   if (op != MPI_SUCCESS) {
     throw std::runtime_error("MPI_Barrier failed, see MPI output for details.");
   }
 }
 
-MPIBroadcast::MPIBroadcast(MPIContext* mpi_context, HorovodGlobalState* global_state)
-    : BroadcastOp(global_state), mpi_context_(mpi_context) {}
+MPIBroadcast::MPIBroadcast(HorovodGlobalState* global_state)
+    : BroadcastOp(global_state) {}
+
+Status MPIBroadcast::Execute(std::vector<TensorTableEntry>& entries,
+                             const Response& response) {
+  WaitForData(entries);
 
-Status MPIBroadcast::Execute(std::vector<TensorTableEntry>& entries, const Response& response) {
   assert(entries.size() == 1);
   auto e = entries[0];
+  auto& process_set = global_state_->process_set_table.Get(e.process_set_id);
+  const auto& mpi_context = process_set.mpi_context;
 
   // On root rank, MPI_Bcast sends data, on other ranks it receives data.
   void* data_ptr;
-  if (global_state_->controller->GetRank() == e.root_rank) {
+  if (process_set.controller->GetRank() == e.root_rank) {
     data_ptr = (void*) e.tensor->data();
   } else {
     data_ptr = (void*) e.output->data();
@@ -369,9 +394,9 @@ Status MPIBroadcast::Execute(std::vector<TensorTableEntry>& entries, const Respo
   global_state_->timeline.ActivityStartAll(entries, MPI_BCAST);
   int op = MPI_Bcast(data_ptr,
                      (int) e.tensor->shape().num_elements(),
-                     mpi_context_->GetMPIDataType(e.tensor->dtype()),
+                     mpi_context.GetMPIDataType(e.tensor->dtype()),
                      e.root_rank,
-                     mpi_context_->GetMPICommunicator(Communicator::GLOBAL));
+                     mpi_context.GetMPICommunicator(Communicator::GLOBAL));
   if (op != MPI_SUCCESS) {
     throw std::runtime_error("MPI_Broadcast failed, see MPI output for details.");
   }
@@ -386,12 +411,16 @@ bool MPIBroadcast::Enabled(const ParameterManager& param_manager,
   return true;
 }
 
-MPIAlltoall::MPIAlltoall(MPIContext* mpi_context, HorovodGlobalState* global_state)
-    : AlltoallOp(global_state), mpi_context_(mpi_context) {}
+MPIAlltoall::MPIAlltoall(HorovodGlobalState* global_state)
+    : AlltoallOp(global_state) {}
 
 Status MPIAlltoall::Execute(std::vector<TensorTableEntry>& entries, const Response& response) {
+  WaitForData(entries);
+
   assert(entries.size() == 1);
   auto e = entries[0];
+  auto& process_set = global_state_->process_set_table.Get(e.process_set_id);
+  const auto& mpi_context = process_set.mpi_context;
 
   std::vector<int32_t> sdispls, rdispls;
   std::vector<int32_t> sendcounts, recvcounts;
@@ -405,10 +434,10 @@ Status MPIAlltoall::Execute(std::vector<TensorTableEntry>& entries, const Respon
   global_state_->timeline.ActivityStartAll(entries, MPI_ALLTOALL);
 
   int op = MPI_Alltoallv(sendbuf, sendcounts.data(), sdispls.data(),
-                         mpi_context_->GetMPIDataType(e.tensor->dtype()),
+                         mpi_context.GetMPIDataType(e.tensor->dtype()),
                          buffer_data, recvcounts.data(), rdispls.data(),
-                         mpi_context_->GetMPIDataType(e.output->dtype()),
-                         mpi_context_->GetMPICommunicator(Communicator::GLOBAL));
+                         mpi_context.GetMPIDataType(e.output->dtype()),
+                         mpi_context.GetMPICommunicator(Communicator::GLOBAL));
   if (op != MPI_SUCCESS) {
     throw std::runtime_error("MPI_Alltoallv failed, see MPI output for details.");
   }
@@ -423,5 +452,74 @@ bool MPIAlltoall::Enabled(const ParameterManager& param_manager,
   return true;
 }
 
+MPIReducescatter::MPIReducescatter(HorovodGlobalState* global_state)
+    : ReducescatterOp(global_state) {}
+
+Status MPIReducescatter::Execute(std::vector<TensorTableEntry>& entries,
+                                 const Response& response) {
+  assert(!entries.empty());
+  WaitForData(entries);
+
+  auto& first_entry = entries[0];
+  auto& process_set =
+      global_state_->process_set_table.Get(first_entry.process_set_id);
+  const auto& mpi_context = process_set.mpi_context;
+
+  auto& timeline = global_state_->timeline;
+
+  const void* sendbuf = nullptr;
+  void* buffer_data = nullptr;
+  int global_rank = process_set.controller->GetRank();
+  int global_size = process_set.controller->GetSize();
+  auto output_shapes = ComputeOutputShapes(entries, global_size);
+  std::vector<int> recvcounts = ComputeReceiveCounts(output_shapes);
+
+  timeline.ActivityStartAll(entries, ALLOCATE_OUTPUT);
+  Status status = AllocateOutput(entries, output_shapes[global_rank]);
+  if (!status.ok()) {
+    return status;
+  }
+  timeline.ActivityEndAll(entries);
+
+  // Copy memory into the fusion buffer.
+  if (entries.size() > 1) {
+    timeline.ActivityStartAll(entries, MEMCPY_IN_FUSION_BUFFER);
+    int element_size = mpi_context.GetMPITypeSize(first_entry.tensor->dtype());
+    MemcpyInFusionBuffer(entries, output_shapes, element_size, buffer_data);
+    timeline.ActivityEndAll(entries);
+  } else {
+    sendbuf = first_entry.tensor->data();
+    buffer_data = (void*)first_entry.output->data();
+  }
+
+  // Do reducescatter.
+  timeline.ActivityStartAll(entries, MPI_REDUCESCATTER);
+  int op = MPI_Reduce_scatter(
+      sendbuf != nullptr ? sendbuf : MPI_IN_PLACE, buffer_data,
+      recvcounts.data(), mpi_context.GetMPIDataType(first_entry.tensor),
+      mpi_context.GetMPISumOp(first_entry.tensor->dtype()),
+      mpi_context.GetMPICommunicator(Communicator::GLOBAL));
+  if (op != MPI_SUCCESS) {
+    throw std::runtime_error(
+        "MPI_Reduce_scatter failed, see MPI output for details.");
+  }
+  timeline.ActivityEndAll(entries);
+
+  // Copy memory out of the fusion buffer.
+  if (entries.size() > 1) {
+    timeline.ActivityStartAll(entries, MEMCPY_OUT_FUSION_BUFFER);
+    MemcpyOutFusionBuffer(buffer_data, entries);
+    timeline.ActivityEndAll(entries);
+  }
+
+  return Status::OK();
+}
+
+bool MPIReducescatter::Enabled(const ParameterManager& param_manager,
+                               const std::vector<TensorTableEntry>& entries,
+                               const Response& response) const {
+  return true;
+}
+
 } // namespace common
 } // namespace horovod
