@@ -219,12 +219,12 @@ def get_average_backwards_compatibility_fun(reduce_ops):
     hvd.allreduce(tensor, average=False, op=hvd.Adasum)).
     """
     def impl(op, average):
-        if op != None:
-            if average != None:
+        if op is not None:
+            if average is not None:
                 raise ValueError('The op parameter supersedes average. Please provide only one of them.')
             return op
-        elif average != None:
-            warnings.warn('Parameter `average` has been replaced with `op` and will be removed in v0.21.0',
+        elif average is not None:
+            warnings.warn('Parameter `average` has been replaced with `op` and will be removed in v1.0',
                           DeprecationWarning)
             return reduce_ops.Average if average else reduce_ops.Sum
         else:
