@@ -136,6 +136,10 @@ extern "C" int horovod_mxnet_alltoall_async(NDArray* input,
                                             NDArray* output_received_splits,
                                             int priority,
                                             int process_set_id);
+extern "C" int
+horovod_mxnet_reducescatter_async(NDArray* input, NDArray* output,
+                                  const char* name, int priority,
+                                  int process_set_id);
 
 } // namespace mxnet
 } // namespace horovod
