@@ -12,11 +12,12 @@
 # See the License for the specific language governing permissions and
 # limitations under the License.
 # =============================================================================
+import warnings
 
 
 class _HorovodArgs(object):
     def __init__(self):
-        self.np = 1
+        self.num_proc = 1
         self.check_build = None
         self.ssh_port = None
         self.ssh_identity_file = None
@@ -49,8 +50,8 @@ class _HorovodArgs(object):
         self.autotune_gaussian_process_noise = None
 
         # elastic arguments
-        self.min_np = None
-        self.max_np = None
+        self.min_num_proc = None
+        self.max_num_proc = None
         self.slots = None
         self.elastic_timeout = None
         self.reset_limit = None
@@ -94,14 +95,15 @@ def run(
         func,
         args=(),
         kwargs=None,
-        np=1,
-        min_np=None,
-        max_np=None,
+        num_proc=1,
+        min_num_proc=None,
+        max_num_proc=None,
         slots=None,
         reset_limit=None,
         cooldown_range=None,
         hosts=None,
         hostfile=None,
+        host_discovery_script=None,
         start_timeout=None,
         ssh_port=None,
         ssh_identity_file=None,
@@ -111,8 +113,16 @@ def run(
         use_gloo=None,
         use_mpi=None,
         mpi_args=None,
+        # network_interface is deprecated, use network_interfaces instead
         network_interface=None,
-        executable=None):
+        network_interfaces=None,
+        executable=None,
+        # np is deprecated, use num_proc instead
+        np=None,
+        # min_np is deprecated, use min_num_proc instead
+        min_np=None,
+        # max_np is deprecated, use max_num_proc instead
+        max_np=None):
     """
     Launch a Horovod job to run the specified process function and get the return value.
 
@@ -121,11 +131,11 @@ def run(
                  This function must be compatible with pickle.
     :param args: Arguments to pass to `func`.
     :param kwargs: Keyword arguments to pass to `func`.
-    :param np: Number of Horovod processes.
-    :param min_np: Minimum number of processes running for training to continue. If number of
+    :param num_proc: Number of Horovod processes.
+    :param min_num_proc: Minimum number of processes running for training to continue. If number of
                    available processes dips below this threshold, then training will wait for
-                   more instances to become available. Defaults to np
-    :param max_np: Maximum number of training processes, beyond which no additional processes
+                   more instances to become available. Defaults to num_proc
+    :param max_num_proc: Maximum number of training processes, beyond which no additional processes
                    will be created. If not specified, then will be unbounded.
     :param slots: Number of slots for processes per host. Normally 1 slot per GPU per host.
                   If slots are provided by the output of the host discovery script, then that
@@ -140,10 +150,17 @@ def run(
     :param hosts: List of host names and the number of available slots
                   for running processes on each, of the form: <hostname>:<slots>
                   (e.g.: host1:2,host2:4,host3:1 indicating 2 processes can run on host1,
-                  4 on host2, and 1 on host3). If not specified, defaults to using localhost:<np>
+                  4 on host2, and 1 on host3). If not specified, defaults to using localhost:<num_proc>
     :param hostfile: Path to a host file containing the list of host names and the number of
                      available slots. Each line of the file must be of the form:
                      <hostname> slots=<slots>
+    :param host_discovery_script: Used for elastic training (autoscaling and fault tolerance).
+                                  An executable script that will print to stdout every available host (one per
+                                  newline character) that can be used to run worker processes. Optionally
+                                  specifies number of slots on the same line as the hostname as: "hostname:slots".
+                                  Providing a discovery script enables elastic training.
+                                  The job will fail immediately if execution of the script returns a non-zero exit
+                                  code on the first call. Subsequent calls will be retried until timeout.
     :param start_timeout: Horovodrun has to perform all the checks and
                           start the processes before the specified
                           timeout. The default value is 30 seconds.
@@ -167,13 +184,27 @@ def run(
     :param use_mpi: Run Horovod using the MPI controller. This will
                     be the default if Horovod was built with MPI support.
     :param mpi_args: Extra arguments for the MPI controller. This is only used when use_mpi is True.
-    :param network_interface: Network interfaces to use for communication separated by comma. If
-                             not specified, Horovod will find the common NICs among all the
-                             workers and use those; example, eth0,eth1.
+    :param network_interfaces: List of network interfaces to use for communication. If not specified,
+                               Horovod will find the common NICs among all the workers.
+                               Example: ["eth0", "eth1"].
     :param executable: Optional executable to run when launching the workers. Defaults to `sys.executable`.
     :return: Return a list which contains values return by all Horovod processes.
              The index of the list corresponds to the rank of each Horovod process.
+             Returns only the first min_num_proc results, if set.
     """
+    if network_interface:
+        network_interfaces = network_interface.split(',')
+        warnings.warn('network_interface is deprecated, use network_interfaces instead', DeprecationWarning)
+    if np is not None:
+        num_proc = np
+        warnings.warn('np is deprecated, use num_proc instead', DeprecationWarning)
+    if min_np is not None:
+        min_num_proc = min_np
+        warnings.warn('min_np is deprecated, use min_num_proc instead', DeprecationWarning)
+    if max_np is not None:
+        max_num_proc = max_np
+        warnings.warn('max_np is deprecated, use max_num_proc instead', DeprecationWarning)
+
     from .launch import _run
 
     if kwargs is None:
@@ -190,14 +221,15 @@ def run(
 
     hargs = _HorovodArgs()
 
-    hargs.np = np
-    hargs.min_np = min_np
-    hargs.max_np = max_np
+    hargs.num_proc = num_proc
+    hargs.min_num_proc = min_num_proc
+    hargs.max_num_proc = max_num_proc
     hargs.slots = slots
     hargs.reset_limit = reset_limit
     hargs.cooldown_range = cooldown_range
     hargs.hosts = hosts
     hargs.hostfile = hostfile
+    hargs.host_discovery_script = host_discovery_script
     hargs.start_timeout = start_timeout
     hargs.ssh_port = ssh_port
     hargs.ssh_identity_file = ssh_identity_file
@@ -207,7 +239,7 @@ def run(
     hargs.verbose = verbose
     hargs.use_gloo = use_gloo
     hargs.use_mpi = use_mpi
-    hargs.nics = network_interface
+    hargs.nics = set(network_interfaces) if network_interfaces else None
     hargs.run_func = wrapped_func
     hargs.executable = executable
 
