@@ -68,6 +68,10 @@ class EstimatorParams(Params):
                                        'number of steps (batches) for validation per epoch',
                                        typeConverter=TypeConverters.toInt)
 
+    random_seed = Param(Params._dummy(), 'random_seed',
+                        'random seed to use for DL frameworks',
+                        typeConverter=TypeConverters.toInt)
+
     shuffle_buffer_size = Param(Params._dummy(),
                                 'shuffle_buffer_size',
                                 'shuffling buffer size of data before training in number of samples',
@@ -92,6 +96,25 @@ class EstimatorParams(Params):
 
     label_shapes = Param(Params._dummy(), 'label_shapes', 'specifies the shape (or shapes) of the label column (or columns)')
 
+    inmemory_cache_all = Param(Params._dummy(), 'inmemory_cache_all',
+                               'Cache the data in memory for training and validation.',
+                               typeConverter=TypeConverters.toBoolean)
+
+    use_gpu = Param(Params._dummy(), 'use_gpu',
+                    'Whether to use the GPU for training. '
+                    'Setting this to False will skip binding to GPU even when GPU is available. '
+                    'Defaults to True.',
+                    typeConverter=TypeConverters.toBoolean)
+
+    mp_start_method = Param(Params._dummy(), 'mp_start_method',
+                    'The start method of multiprocessing. '
+                    'This controls the start up method of python multiprocessing. On unix-based systems, acceptable values are '
+                    '["fork", "spawn", "forkserver"]. Only "Spawn" is accepted on Windows. '
+                    'If not set, default values will be internally set by python, please refer to: '
+                    'https://docs.python.org/3/library/multiprocessing.html#multiprocessing.get_start_method.'
+                    'This param defaults to None.',
+                    typeConverter=TypeConverters.toString)
+
     def __init__(self):
         super(EstimatorParams, self).__init__()
 
@@ -115,6 +138,7 @@ class EstimatorParams(Params):
             epochs=1,
             verbose=1,
             callbacks=[],
+            random_seed=None,
             shuffle_buffer_size=None,
             partitions_per_process=10,
             run_id=None,
@@ -124,7 +148,10 @@ class EstimatorParams(Params):
             train_reader_num_workers=2,
             val_reader_num_workers=2,
             reader_pool_type='process',
-            label_shapes=None)
+            label_shapes=None,
+            inmemory_cache_all=False,
+            use_gpu=True,
+            mp_start_method=None)
 
     def _check_params(self, metadata):
         model = self.getModel()
@@ -269,6 +296,12 @@ class EstimatorParams(Params):
     def getCompressSparseCols(self):
         return self.getOrDefault(self.compress_sparse_cols)
 
+    def setRandomSeed(self, value):
+        return self._set(random_seed=value)
+
+    def getRandomSeed(self):
+        return self.getOrDefault(self.random_seed)
+
     def setShufflingBufferSize(self, value):
         return self._set(shuffle_buffer_size=value)
 
@@ -323,6 +356,23 @@ class EstimatorParams(Params):
     def getLabelShapes(self):
         return self.getOrDefault(self.label_shapes)
 
+    def setInMemoryCacheAll(self, value):
+        return self._set(inmemory_cache_all=value)
+
+    def getInMemoryCacheAll(self):
+        return self.getOrDefault(self.inmemory_cache_all)
+
+    def setUseGpu(self, value):
+        self._set(use_gpu=value)
+
+    def getUseGpu(self):
+        return self.getOrDefault(self.use_gpu)
+
+    def setMpStartMethod(self, value):
+        self._set(mp_start_method=value)
+
+    def getMpStartMethod(self):
+        return self.getOrDefault(self.mp_start_method)
 
 class ModelParams(HasOutputCols):
     history = Param(Params._dummy(), 'history', 'history')
