@@ -25,7 +25,7 @@ from distutils.version import LooseVersion
 
 from horovod.spark.common import constants
 from horovod.spark.common.store import DBFSLocalStore
-from horovod.spark.common.util import _get_assigned_gpu_or_default
+from horovod.spark.common.util import _get_assigned_gpu_or_default, _set_mp_start_method
 from horovod.runner.common.util import codec
 
 
@@ -47,10 +47,13 @@ def RemoteTrainer(estimator, metadata, keras_utils, run_id, dataset_idx):
     sample_weight_col = estimator.getSampleWeightCol()
     custom_objects = estimator.getCustomObjects()
     should_validate = estimator.getValidation()
+    random_seed = estimator.getRandomSeed()
     user_shuffle_buffer_size = estimator.getShufflingBufferSize()
     user_verbose = estimator.getVerbose()
     checkpoint_callback = estimator.getCheckpointCallback()
     inmemory_cache_all = estimator.getInMemoryCacheAll()
+    should_use_gpu = estimator.getUseGpu()
+    mp_start_method = estimator.getMpStartMethod()
 
     # Data reader parameters
     train_reader_worker_count = estimator.getTrainReaderNumWorker()
@@ -102,6 +105,10 @@ def RemoteTrainer(estimator, metadata, keras_utils, run_id, dataset_idx):
         yield None
 
     def train(serialized_model, train_rows, val_rows, avg_row_size):
+        # If not empty, set it before everything else.
+        if mp_start_method:
+            _set_mp_start_method(mp_start_method, user_verbose)
+
         from petastorm import TransformSpec, make_reader, make_batch_reader
         import horovod as _horovod
         k = get_keras()
@@ -110,7 +117,22 @@ def RemoteTrainer(estimator, metadata, keras_utils, run_id, dataset_idx):
         hvd = get_horovod()
         hvd.init()
 
-        pin_gpu(hvd, tf, k)
+        # Verbose mode 1 will print a progress bar
+        verbose = user_verbose if hvd.rank() == 0 else 0
+
+        if should_use_gpu:
+            if verbose:
+                print("Pinning current process to the GPU.")
+            pin_gpu(hvd, tf, k)
+        else:
+            if verbose:
+                print("Skip pinning current process to the GPU.")
+
+        if random_seed is not None:
+            if LooseVersion(tf.__version__) < LooseVersion('2.0.0'):
+                tf.random.set_random_seed(random_seed)
+            else:
+                tf.random.set_seed(random_seed)
 
         # If user specifies any user_shuffle_buffer_size (even 0), we should honor it.
         if user_shuffle_buffer_size is None:
@@ -130,8 +152,6 @@ def RemoteTrainer(estimator, metadata, keras_utils, run_id, dataset_idx):
         scaled_lr = k.backend.get_value(model.optimizer.lr) * hvd.size()
         k.backend.set_value(model.optimizer.lr, scaled_lr)
 
-        # Verbose mode 1 will print a progress bar
-        verbose = user_verbose if hvd.rank() == 0 else 0
 
         if verbose:
             print(f"Shared lib path is pointing to: {_horovod.common.process_sets._basics.MPI_LIB_CTYPES}")
@@ -217,7 +237,8 @@ def RemoteTrainer(estimator, metadata, keras_utils, run_id, dataset_idx):
                 schema_fields.append(sample_weight_col)
 
             if verbose:
-                print(f"Training parameters: Epochs: {epochs}, Scaled lr: {scaled_lr}, Shuffle size: {shuffle_buffer_size}\n"
+                print(f"Training parameters: Epochs: {epochs}, Scaled lr: {scaled_lr}, "
+                      f"Shuffle size: {shuffle_buffer_size}, random_seed: {random_seed}\n"
                       f"Train rows: {train_rows}, Train batch size: {batch_size}, Train_steps_per_epoch: {steps_per_epoch}\n"
                       f"Val rows: {val_rows}, Val batch size: {val_batch_size}, Val_steps_per_epoch: {validation_steps}\n"
                       f"Checkpoint file: {remote_store.checkpoint_path}, Logs dir: {remote_store.logs_path}\n")
@@ -263,7 +284,7 @@ def RemoteTrainer(estimator, metadata, keras_utils, run_id, dataset_idx):
 
                     train_data = make_dataset(train_reader, batch_size, shuffle_buffer_size,
                                               is_batch_reader, shuffle=True if shuffle_buffer_size > 0 else False,
-                                              cache=inmemory_cache_all)
+                                              cache=inmemory_cache_all, seed=random_seed)
                     val_data = make_dataset(val_reader, val_batch_size, shuffle_buffer_size,
                                             is_batch_reader, shuffle=False, cache=inmemory_cache_all) \
                         if val_reader else None
