@@ -23,7 +23,7 @@ import torch
 from torch.utils.tensorboard import SummaryWriter
 
 from horovod.spark.common import constants
-from horovod.spark.common.util import _get_assigned_gpu_or_default, to_list
+from horovod.spark.common.util import _get_assigned_gpu_or_default, to_list, _set_mp_start_method
 from horovod.spark.common.store import DBFSLocalStore
 from horovod.spark.torch.util import deserialize_fn
 
@@ -50,6 +50,7 @@ def RemoteTrainer(estimator, metadata, last_checkpoint_state, run_id, dataset_id
     validation_steps_per_epoch = estimator.getValidationStepsPerEpoch()
     sample_weight_col = estimator.getSampleWeightCol()
     metric_fn_groups = estimator.getMetrics()
+    random_seed = estimator.getRandomSeed()
     user_shuffle_buffer_size = estimator.getShufflingBufferSize()
     user_verbose = estimator.getVerbose()
     train_minibatch_fn = estimator.getTrainMinibatchFn()
@@ -59,6 +60,8 @@ def RemoteTrainer(estimator, metadata, last_checkpoint_state, run_id, dataset_id
     transformation_fn = estimator.getTransformationFn()
     transformation = transformation_fn if transformation_fn else None
     inmemory_cache_all = estimator.getInMemoryCacheAll()
+    should_use_gpu = estimator.getUseGpu()
+    mp_start_method = estimator.getMpStartMethod()
 
     # If loss weight is not provided, use equal loss for all the labels
     loss_weights = estimator.getLossWeights()
@@ -91,6 +94,7 @@ def RemoteTrainer(estimator, metadata, last_checkpoint_state, run_id, dataset_id
     store = estimator.getStore()
     remote_store = store.to_remote(run_id, dataset_idx)
     is_dbfs = isinstance(store, DBFSLocalStore)
+    storage_options = store.storage_options
 
     @contextlib.contextmanager
     def empty_batch_reader():
@@ -98,11 +102,18 @@ def RemoteTrainer(estimator, metadata, last_checkpoint_state, run_id, dataset_id
 
     def train(serialized_model, optimizer_cls, model_opt_state_serialized,
               train_rows, val_rows, avg_row_size):
+        # If not empty, set it before everything else.
+        if mp_start_method:
+            _set_mp_start_method(mp_start_method, user_verbose)
+
         from petastorm import TransformSpec, make_reader, make_batch_reader
         from petastorm.pytorch import BatchedDataLoader, InMemBatchedDataLoader
         import torch
         import horovod.torch as hvd
 
+        if random_seed is not None:
+            torch.manual_seed(random_seed)
+
         # Deserializing objects
         model_opt_state = torch.load(model_opt_state_serialized)
         model = deserialize(serialized_model)
@@ -116,13 +127,35 @@ def RemoteTrainer(estimator, metadata, last_checkpoint_state, run_id, dataset_id
         # Horovod: initialize library.
         hvd.init()
 
-        if not user_shuffle_buffer_size:
+        if user_verbose:
+            import horovod as _horovod
+            print(f"Shared lib path is pointing to: {_horovod.common.process_sets._basics.MPI_LIB_CTYPES}")
+
+        # If user specifies any user_shuffle_buffer_size (even 0), we should honor it.
+        if user_shuffle_buffer_size is None:
             shuffle_buffer_size = \
                 calculate_shuffle_buffer_size(hvd, avg_row_size, train_rows / hvd.size())
         else:
+            if user_shuffle_buffer_size < 0:
+                raise ValueError("user_shuffle_buffer_size cannot be negative!")
             shuffle_buffer_size = user_shuffle_buffer_size
 
+        if not should_use_gpu and user_verbose:
+            print("Skip pinning current process to the GPU.")
+
         cuda_available = torch.cuda.is_available()
+
+        if cuda_available and not should_use_gpu:
+            print("GPU is available but use_gpu is set to False."
+                  "Training will proceed without GPU support.")
+            cuda_available = False
+
+        # We need to check all ranks have same device type for traning.
+        # Horovod doesn't support heterogeneous allreduce for gradients.
+        cuda_avail_list = hvd.allgather_object(cuda_available, name='device type')
+        if cuda_avail_list.count(cuda_available) != hvd.size():
+            raise RuntimeError("All ranks don't have same device type!")
+
         if cuda_available:
             # Horovod: pin GPU to local rank or the assigned GPU from spark.
             torch.cuda.set_device(_get_assigned_gpu_or_default(default=hvd.local_rank()))
@@ -204,7 +237,11 @@ def RemoteTrainer(estimator, metadata, last_checkpoint_state, run_id, dataset_id
                 if cuda_available:
                     model.cuda()
 
-
+            if hvd.rank() == 0 and user_verbose:
+                print(f"Training parameters: Epochs: {epochs}\n"
+                      f"Train rows: {train_rows}, Train batch size: {batch_size}, Train_steps_per_epoch: {steps_per_epoch}\n"
+                      f"Shuffle buffer size: {shuffle_buffer_size}, Random seed: {random_seed}\n"
+                      f"Checkpoint file: {ckpt_file}, Logs dir: {logs_dir}\n")
             # In general, make_batch_reader is faster than make_reader for reading the dataset.
             # However, we found out that make_reader performs data transformations much faster than
             # make_batch_reader with parallel worker processes. Therefore, the default reader
@@ -230,6 +267,9 @@ def RemoteTrainer(estimator, metadata, last_checkpoint_state, run_id, dataset_id
                                 hdfs_driver=PETASTORM_HDFS_DRIVER,
                                 schema_fields=schema_fields,
                                 transform_spec=transform_spec,
+                                storage_options=storage_options,
+                                # Don't shuffle row groups without shuffling.
+                                shuffle_row_groups=True if shuffle_buffer_size > 0 else False,
                                 **reader_factory_kwargs) as train_reader:
                 with reader_factory(remote_store.val_data_path,
                                     num_epochs=None,
@@ -240,6 +280,8 @@ def RemoteTrainer(estimator, metadata, last_checkpoint_state, run_id, dataset_id
                                     hdfs_driver=PETASTORM_HDFS_DRIVER,
                                     schema_fields=schema_fields,
                                     transform_spec=transform_spec,
+                                    storage_options=storage_options,
+                                    shuffle_row_groups=False,
                                     **reader_factory_kwargs) \
                     if should_validate else empty_batch_reader() as val_reader:
 
@@ -342,12 +384,15 @@ def RemoteTrainer(estimator, metadata, last_checkpoint_state, run_id, dataset_id
                         else:
                             validation_steps = validation_steps_per_epoch
 
+                        if hvd.rank() == 0 and user_verbose:
+                            print(f"Val rows: {val_rows}, Val batch size: {val_batch_size}, Val_steps_per_epoch: {validation_steps}\n")
+
                         if inmemory_cache_all:
                             # Petastorm introduced InMemBatchedDataLoader class in v0.11.0
                             val_loader = InMemBatchedDataLoader(val_reader,
                                                                 batch_size=val_batch_size,
                                                                 num_epochs=epochs,
-                                                                rows_capacity=validation_steps_per_epoch*val_batch_size,
+                                                                rows_capacity=validation_steps*val_batch_size,
                                                                 shuffle=False)
                         else:
                             val_loader = BatchedDataLoader(val_reader,
