@@ -41,7 +41,6 @@ class LocalGradientAggregationHelperEager:
         # is equal to 0.
         self.counter = tf.Variable(initial_value=0)
 
-    @tf.function
     def compute_gradients(self, grads, vars):
         # On steps where allreduce happens, resulting_grads returns the allreduced
         # gradients, on other steps it returns the locally aggregated
@@ -120,8 +119,8 @@ class LocalGradientAggregationHelperEager:
     def _clear_vars(self):
         self.counter.assign(0)
         for idx in self.locally_aggregated_grads.keys():
-            self.locally_aggregated_grads[idx].assign_add(
-                -1 * self.locally_aggregated_grads[idx])
+            self.locally_aggregated_grads[idx].assign(
+                tf.zeros_like(self.locally_aggregated_grads[idx]))
 
     def apply_gradients(self, apply_grads_closure, optimizer, *args, **kwargs):
         def increment_optimizer_iteration():
@@ -141,11 +140,6 @@ class LocalGradientAggregationHelperEager:
             return increment_optimizer_iteration()
 
         def is_aggregation_step():
-            if _POST_TF_2_4_0:
-                # In TF 2.4+ we evaluate whether it's time to aggregated gradients before
-                # calling `_aggregate_gradients()`.
-                return tf.equal(tf.add(self.counter, 1), self.backward_passes_per_step)
-
             return tf.equal(self.counter, 0)
 
         return tf.cond(
