@@ -27,6 +27,10 @@
 #define EIGEN_USE_GPU
 #endif  // HAVE_CUDA || HAVE_ROCM
 
+#if HAVE_ROCM
+#define EIGEN_USE_HIP
+#endif
+
 #include "tensorflow/core/framework/op.h"
 #include "tensorflow/core/framework/op_kernel.h"
 #include "tensorflow/core/framework/shape_inference.h"
@@ -186,10 +190,12 @@ public:
       int64_t size, std::shared_ptr<common::PersistentBuffer>* tensor) override;
   virtual common::Status
   AllocateOutput(common::TensorShape shape,
-                 std::shared_ptr<common::Tensor>* tensor) override;
+                 std::shared_ptr<common::Tensor>* tensor,
+                 std::shared_ptr<common::ReadyEvent>* event = nullptr) override;
   virtual common::Status
   AllocateOutput(int output_index, common::TensorShape shape,
-                 std::shared_ptr<common::Tensor>* tensor) override;
+                 std::shared_ptr<common::Tensor>* tensor,
+                 std::shared_ptr<common::ReadyEvent>* event = nullptr) override;
   virtual common::Status
   AllocateZeros(int64_t num_elements, common::DataType dtype,
                 std::shared_ptr<common::Tensor>* tensor) override;
@@ -303,6 +309,18 @@ int64_t TFTensor::size() const { return (int64_t)tensor_.tensor_data().size(); }
 
 const ::tensorflow::Tensor*  TFTensor::tensor() const { return &tensor_; }
 
+// On GPU this event will signal that data is ready, and tensors are
+// allocated.
+#if HAVE_GPU
+common::ReadyEvent* RecordReadyEvent(OpKernelContext* context) {
+  auto device_context = context->op_device_context();
+  if (device_context != nullptr) {
+    return new TFReadyEvent(context);
+  }
+  return nullptr;
+}
+#endif
+
 TFOpContext::TFOpContext(OpKernelContext* context) : context_(context) {}
 
 common::Status TFOpContext::AllocatePersistent(
@@ -317,13 +335,15 @@ common::Status TFOpContext::AllocatePersistent(
 
 common::Status
 TFOpContext::AllocateOutput(common::TensorShape shape,
-                            std::shared_ptr<common::Tensor>* tensor) {
-  return TFOpContext::AllocateOutput(0, shape, tensor);
+                            std::shared_ptr<common::Tensor>* tensor,
+                            std::shared_ptr<common::ReadyEvent>* event) {
+  return TFOpContext::AllocateOutput(0, shape, tensor, event);
 }
 
 common::Status
 TFOpContext::AllocateOutput(int output_index, common::TensorShape shape,
-                            std::shared_ptr<common::Tensor>* tensor) {
+                            std::shared_ptr<common::Tensor>* tensor,
+                            std::shared_ptr<common::ReadyEvent>* event) {
   TensorShape tf_shape;
   for (int idx = 0; idx < shape.dims(); ++idx) {
     tf_shape.AddDim(shape.dim_size(idx));
@@ -338,7 +358,11 @@ TFOpContext::AllocateOutput(int output_index, common::TensorShape shape,
   // complete.
   auto device_context = context_->op_device_context();
   if (device_context != nullptr) {
-    device_context->stream()->BlockHostUntilDone();
+    if (event == nullptr) {
+      device_context->stream()->BlockHostUntilDone();
+    } else {
+      *event = std::shared_ptr<common::ReadyEvent>(RecordReadyEvent(context_));
+    }
   }
 #endif
   return ConvertStatus(status);
@@ -396,25 +420,20 @@ OpKernelContext* TFOpContext::GetKernelContext() const { return context_; }
 
 int GetDeviceID(OpKernelContext* context) {
   int device = CPU_DEVICE_ID;
+#if TENSORFLOW_VERSION >= 2009000000
+  if (context->device() != nullptr &&
+      context->device()->tensorflow_accelerator_device_info() != nullptr) {
+    device = context->device()->tensorflow_accelerator_device_info()->gpu_id;
+  }
+#else
   if (context->device() != nullptr &&
       context->device()->tensorflow_gpu_device_info() != nullptr) {
     device = context->device()->tensorflow_gpu_device_info()->gpu_id;
   }
+#endif // TENSORFLOW_VERSION >= 2009000000
   return device;
 }
 
-// On GPU this event will signal that data is ready, and tensors are
-// allocated.
-#if HAVE_GPU
-common::ReadyEvent* RecordReadyEvent(OpKernelContext* context) {
-  auto device_context = context->op_device_context();
-  if (device_context != nullptr) {
-    return new TFReadyEvent(context);
-  }
-  return nullptr;
-}
-#endif
-
 } // namespace
 
 class HorovodAllreduceOp : public AsyncOpKernel {
@@ -504,7 +523,7 @@ REGISTER_OP("HorovodAllreduce")
       return Status::OK();
     })
     .Doc(R"doc(
-Perform an MPI Allreduce on a tensor. All other processes that do a reduction
+Perform an Allreduce on a tensor. All other processes that do a reduction
 on a tensor with the same name must have the same dimension for that tensor.
 Tensors are reduced with other tensors that have the same node name for the
 allreduce.
@@ -513,7 +532,7 @@ Arguments
     tensor:     A tensor to reduce.
 
 Output
-    sum:    A tensor with the same shape as `tensor`, summed across all MPI processes.
+    sum:    A tensor with the same shape as `tensor`, summed across all processes.
 )doc");
 
 class HorovodGroupedAllreduceOp : public AsyncOpKernel {
@@ -733,7 +752,7 @@ REGISTER_OP("HorovodAllgather")
       return Status::OK();
     })
     .Doc(R"doc(
-Perform an MPI Allgather on a tensor. All other processes that do a gather on a
+Perform an Allgather on a tensor. All other processes that do a gather on a
 tensor with the same name must have the same rank for that tensor, and have the
 same dimension on all but the first dimension.
 
@@ -741,7 +760,7 @@ Arguments
     tensor:     A tensor to gather.
 
 Output
-    gathered:    A tensor with the same shape as `tensor` except for the first dimension.
+    output:     A tensor with the same shape as `tensor` except for the first dimension.
 )doc");
 
 class HorovodBroadcastOp : public AsyncOpKernel {
@@ -830,7 +849,7 @@ REGISTER_OP("HorovodBroadcast")
       return Status::OK();
     })
     .Doc(R"doc(
-Perform an MPI Broadcast on a tensor. All other processes that do a broadcast
+Perform a Broadcast on a tensor. All other processes that do a broadcast
 on a tensor with the same name must have the same dimension for that tensor.
 
 Arguments
@@ -1096,6 +1115,105 @@ Input
 )doc");
 #endif // TENSORFLOW_VERSION >= 2006000000
 
+class HorovodReducescatterOp : public AsyncOpKernel {
+public:
+  explicit HorovodReducescatterOp(OpKernelConstruction* context)
+      : AsyncOpKernel(context) {
+    int reduce_op;
+    OP_REQUIRES_OK(context, context->GetAttr("reduce_op", &reduce_op));
+    reduce_op_ = static_cast<horovod::common::ReduceOp>(reduce_op);
+    OP_REQUIRES_OK(context,
+                   context->GetAttr("ignore_name_scope", &ignore_name_scope_));
+    OP_REQUIRES_OK(context,
+                   context->GetAttr("process_set_id", &process_set_id_));
+  }
+
+  void ComputeAsync(OpKernelContext* context, DoneCallback done) override {
+    OP_REQUIRES_OK_ASYNC(context, ConvertStatus(common::CheckInitialized()),
+                         done);
+
+    auto node_name = name();
+    if (ignore_name_scope_) {
+      auto pos = node_name.find_last_of('/');
+      if (pos != std::string::npos) {
+        node_name = node_name.substr(pos + 1);
+      }
+    }
+    auto device = GetDeviceID(context);
+    auto tensor = context->input(0);
+    // ReadyEvent makes sure input tensor is ready.
+    common::ReadyEventList ready_event_list;
+#if HAVE_GPU
+    ready_event_list.AddReadyEvent(
+        std::shared_ptr<common::ReadyEvent>(RecordReadyEvent(context)));
+#endif
+    // We cannot pre-allocate output for reducescatter, since shape of result is
+    // only known after all ranks make a request.
+    auto hvd_context = std::make_shared<TFOpContext>(context);
+    auto hvd_tensor = std::make_shared<TFTensor>(tensor);
+    auto enqueue_result = EnqueueTensorReducescatter(
+        hvd_context, hvd_tensor, ready_event_list, node_name, device,
+        [context, done](const common::Status& status) {
+#if HAVE_GPU
+          auto hvd_event = status.event;
+          if (hvd_event.event) {
+            auto device_context = context->op_device_context();
+            if (device_context != nullptr) {
+              auto stream = stream_executor::gpu::AsGpuStreamValue(
+                  device_context->stream());
+              HVD_GPU_CHECK(gpuStreamWaitEvent(stream, *(hvd_event.event), 0));
+            }
+          }
+#endif
+          context->SetStatus(ConvertStatus(status));
+          done();
+        },
+        reduce_op_, process_set_id_);
+    OP_REQUIRES_OK_ASYNC(context, ConvertStatus(enqueue_result), done);
+  }
+
+private:
+  horovod::common::ReduceOp reduce_op_;
+  bool ignore_name_scope_;
+  int process_set_id_;
+};
+
+REGISTER_KERNEL_BUILDER(Name("HorovodReducescatter").Device(DEVICE_CPU),
+                        HorovodReducescatterOp);
+#if HOROVOD_GPU_REDUCESCATTER
+REGISTER_KERNEL_BUILDER(Name("HorovodReducescatter").Device(DEVICE_GPU),
+                        HorovodReducescatterOp);
+#endif
+
+REGISTER_OP("HorovodReducescatter")
+    .Attr("T: {int32, int64, float16, float32, float64}")
+    .Attr("reduce_op: int")
+    .Attr("ignore_name_scope: bool = False")
+    .Attr("process_set_id: int = 0")
+    .Input("tensor: T")
+    .Output("output: T")
+    .SetShapeFn([](shape_inference::InferenceContext* c) {
+      shape_inference::ShapeHandle output;
+      TF_RETURN_IF_ERROR(
+          c->ReplaceDim(c->input(0), 0, c->UnknownDim(), &output));
+      c->set_output(0, output);
+      return Status::OK();
+    })
+    .Doc(R"doc(
+Perform a Reducescatter on a tensor. All other processes that do a
+reduce scatter on a tensor with the same name must have the same shape for
+that tensor. Tensors are reduced with other tensors that have the same node
+name for the reducescatter. The output shape is identical to the input
+shape except for the first dimension, which will be divided across the
+different Horovod processes.
+
+Arguments
+    tensor:     A tensor to reduce and scatter.
+
+Output
+    output:     A tensor with the same shape as `tensor` except for the first dimension.
+)doc");
+
 class HorovodJoinOp : public AsyncOpKernel {
 public:
   explicit HorovodJoinOp(OpKernelConstruction* context)
