@@ -26,6 +26,9 @@ from tensorflow.python.platform import resource_loader
 from horovod.common.util import check_installed_version, get_ext_suffix, \
     get_average_backwards_compatibility_fun, gpu_available, num_rank_is_power_2
 from horovod.common.basics import HorovodBasics as _HorovodBasics
+from horovod.common.process_sets import _setup as _setup_process_sets
+from horovod.common.process_sets import ProcessSet, global_process_set, add_process_set, remove_process_set, \
+    _temp_process_set_object
 from horovod.tensorflow.util import _executing_eagerly
 
 
@@ -54,7 +57,6 @@ else:
 _basics = _HorovodBasics(__file__, 'mpi_lib')
 
 # import basic methods
-init = _basics.init
 shutdown = _basics.shutdown
 is_initialized = _basics.is_initialized
 start_timeline = _basics.start_timeline
@@ -81,12 +83,19 @@ Average = _basics.Average
 Sum = _basics.Sum
 Adasum = _basics.Adasum
 
+def init(*args, **kwargs):
+    _basics.init(*args, **kwargs)
+    # Call set up again to make sure the basics is in sync
+    _setup_process_sets(_basics)
+
 is_homogeneous = _basics.is_homogeneous
 
 handle_average_backwards_compatibility = get_average_backwards_compatibility_fun(_basics)
 
 check_num_rank_power_of_2 = num_rank_is_power_2
 
+_setup_process_sets(_basics)
+
 
 # This function will create a default device map which includes all visible devices.
 # Please run this function in a subprocess
@@ -101,7 +110,7 @@ def _normalize_name(name):
 
 
 def _allreduce(tensor, name=None, op=Sum, prescale_factor=1.0, postscale_factor=1.0,
-               ignore_name_scope=False):
+               ignore_name_scope=False, process_set=global_process_set):
     """An op which reduces an input tensor over all the Horovod processes. The
     default reduction is a sum.
 
@@ -118,7 +127,8 @@ def _allreduce(tensor, name=None, op=Sum, prescale_factor=1.0, postscale_factor=
     return MPI_LIB.horovod_allreduce(tensor, name=name, reduce_op=op,
                                      prescale_factor=prescale_factor,
                                      postscale_factor=postscale_factor,
-                                     ignore_name_scope=ignore_name_scope)
+                                     ignore_name_scope=ignore_name_scope,
+                                     process_set_id=process_set.process_set_id)
 
 
 @ops.RegisterGradient('HorovodAllreduce')
@@ -136,13 +146,15 @@ def _allreduce_grad(op, grad):
     prescale_factor = op.get_attr('prescale_factor')
     postscale_factor = op.get_attr('postscale_factor')
     ignore_name_scope = op.get_attr('ignore_name_scope')
+    process_set_id = op.get_attr('process_set_id')
     return _allreduce(grad, op=reduce_op, prescale_factor=prescale_factor,
                       postscale_factor=postscale_factor,
-                      ignore_name_scope=ignore_name_scope)
+                      ignore_name_scope=ignore_name_scope,
+                      process_set=_temp_process_set_object(process_set_id))
 
 
 def _grouped_allreduce(tensors, name=None, op=Sum, prescale_factor=1.0, postscale_factor=1.0,
-                       ignore_name_scope=False):
+                       ignore_name_scope=False, process_set=global_process_set):
     """An op which reduces input tensors over all the Horovod processes. The
     default reduction is a sum.
 
@@ -165,7 +177,8 @@ def _grouped_allreduce(tensors, name=None, op=Sum, prescale_factor=1.0, postscal
     return MPI_LIB.horovod_grouped_allreduce(tensors, name=name, reduce_op=op,
                                              prescale_factor=prescale_factor,
                                              postscale_factor=postscale_factor,
-                                             ignore_name_scope=ignore_name_scope)
+                                             ignore_name_scope=ignore_name_scope,
+                                             process_set_id=process_set.process_set_id)
 
 
 @ops.RegisterGradient('HorovodGroupedAllreduce')
@@ -183,13 +196,15 @@ def _grouped_allreduce_grad(op, *grads):
     prescale_factor = op.get_attr('prescale_factor')
     postscale_factor = op.get_attr('postscale_factor')
     ignore_name_scope = op.get_attr('ignore_name_scope')
+    process_set_id = op.get_attr('process_set_id')
     # TODO(joshr): should this be done as separate allreduce ops?
     return _grouped_allreduce(list(grads), op=reduce_op, prescale_factor=prescale_factor,
-                      postscale_factor=postscale_factor,
-                      ignore_name_scope=ignore_name_scope)
+                              postscale_factor=postscale_factor,
+                              ignore_name_scope=ignore_name_scope,
+                              process_set=_temp_process_set_object(process_set_id))
 
 
-def allgather(tensor, name=None, ignore_name_scope=False):
+def allgather(tensor, name=None, ignore_name_scope=False, process_set=global_process_set):
     """An op which concatenates the input tensor with the same input tensor on
     all other Horovod processes.
 
@@ -206,7 +221,8 @@ def allgather(tensor, name=None, ignore_name_scope=False):
     if name is None and not _executing_eagerly():
         name = 'HorovodAllgather_%s' % _normalize_name(tensor.name)
     return MPI_LIB.horovod_allgather(tensor, name=name,
-                                     ignore_name_scope=ignore_name_scope)
+                                     ignore_name_scope=ignore_name_scope,
+                                     process_set_id=process_set.process_set_id)
 
 
 @ops.RegisterGradient('HorovodAllgather')
@@ -221,7 +237,10 @@ def _allgather_grad(op, grad):
       The gradient with respect to the input of the op.
     """
     ignore_name_scope = op.get_attr('ignore_name_scope')
-    grad = _allreduce(grad, op=Average, ignore_name_scope=ignore_name_scope)
+    process_set_id = op.get_attr('process_set_id')
+    temp_process_set_object = _temp_process_set_object(process_set_id)
+    grad = _allreduce(grad, op=Average, ignore_name_scope=ignore_name_scope,
+                      process_set=temp_process_set_object)
 
     with tf.device('/cpu:0'):
         # Keep the tensor of split sizes on CPU.
@@ -229,14 +248,16 @@ def _allgather_grad(op, grad):
         d = tf.shape(x)
         d = tf.reshape(d[0], [1])
 
-        s = size()
-        d = tf.reshape(allgather(d, ignore_name_scope=ignore_name_scope), [s])
+        s = temp_process_set_object.size()
+        d = tf.reshape(
+            allgather(d, ignore_name_scope=ignore_name_scope, process_set=temp_process_set_object),
+            [s])
 
     splits = tf.split(grad, num_or_size_splits=d, axis=0)
-    return splits[rank()]
+    return splits[temp_process_set_object.rank()]
 
 
-def broadcast(tensor, root_rank, name=None, ignore_name_scope=False):
+def broadcast(tensor, root_rank, name=None, ignore_name_scope=False, process_set=global_process_set):
     """An op which broadcasts the input tensor on root rank to the same input tensor
     on all other Horovod processes.
 
@@ -251,7 +272,8 @@ def broadcast(tensor, root_rank, name=None, ignore_name_scope=False):
     if name is None and not _executing_eagerly():
         name = 'HorovodBroadcast_%s' % _normalize_name(tensor.name)
     return MPI_LIB.horovod_broadcast(tensor, name=name, root_rank=root_rank,
-                                     ignore_name_scope=ignore_name_scope)
+                                     ignore_name_scope=ignore_name_scope,
+                                     process_set_id=process_set.process_set_id)
 
 
 @ops.RegisterGradient('HorovodBroadcast')
@@ -267,14 +289,53 @@ def _broadcast_grad(op, grad):
     """
     root_rank = op.get_attr('root_rank')
     ignore_name_scope = op.get_attr('ignore_name_scope')
+    process_set_id = op.get_attr('process_set_id')
     grad_reduced = _allreduce(grad, op=Average,
-                              ignore_name_scope=ignore_name_scope)
+                              ignore_name_scope=ignore_name_scope,
+                              process_set=_temp_process_set_object(process_set_id))
     if rank() != root_rank:
         return grad_reduced * 0
     return grad_reduced
 
 
-def alltoall(tensor, splits=None, name=None, ignore_name_scope=False):
+def broadcast_(variables, root_rank, name=None, process_set=global_process_set):
+    """An op which broadcasts the input variables from the root rank to the same
+    input variables on all other Horovod processes. The operation is performed
+    in-place.
+
+    The broadcast operation is keyed by the name of the op combined with the names
+    of the variables. The variable type and shape must be the same on all Horovod
+    processes for any given name. The broadcast will not start until all processes
+    are ready to send and receive all variables. In each process all variables need
+    to be located on the same device (CPU or GPU).
+
+    Note: This is only supported with TensorFlow 2.6 or later.
+
+    Returns:
+      The tensor values of the updated `variables` as broadcasted from root rank.
+    """
+    from distutils.version import LooseVersion
+    if LooseVersion(tf.__version__) < LooseVersion('2.6.0'):
+        raise NotImplementedError("In-place broadcasts are only supported with TensorFlow 2.6 or later")
+
+    from tensorflow.python.ops import resource_variable_ops
+    if all(resource_variable_ops.is_resource_variable(var) for var in variables):
+        with tf.control_dependencies(
+            [MPI_LIB.horovod_broadcast_inplace_resource([var.handle for var in variables],
+                                                        variable_names=[var.name for var in variables],
+                                                        name=name,
+                                                        root_rank=root_rank,
+                                                        process_set_id=process_set.process_set_id)]):
+            return [var.read_value() for var in variables]
+    elif all(not resource_variable_ops.is_resource_variable(var) for var in variables):
+        return MPI_LIB.horovod_broadcast_inplace(variables, variable_names=[var.name for var in variables],
+                                                 name=name, root_rank=root_rank,
+                                                 process_set_id=process_set.process_set_id)
+    else:
+        raise ValueError("All variables passed to broadcast_() should be of the same kind, resource or reference")
+
+
+def alltoall(tensor, splits=None, name=None, ignore_name_scope=False, process_set=global_process_set):
     """An op that scatters slices of the input tensor to all other Horovod processes
     and returns a tensor of gathered slices from all other Horovod processes.
 
@@ -308,7 +369,8 @@ def alltoall(tensor, splits=None, name=None, ignore_name_scope=False):
     if name is None and not _executing_eagerly():
         name = 'HorovodAlltoall_%s' % _normalize_name(tensor.name)
     output, rsplits = MPI_LIB.horovod_alltoall(tensor, splits=splits_, name=name,
-                                               ignore_name_scope=ignore_name_scope)
+                                               ignore_name_scope=ignore_name_scope,
+                                               process_set_id=process_set.process_set_id)
     return (output, rsplits) if splits is not None else output
 
 @ops.RegisterGradient('HorovodAlltoall')
@@ -324,18 +386,71 @@ def _alltoall_grad(op, grad_wrt_output, grad_wrt_received_splits):
       The gradient with respect to the input of the op.
     """
     ignore_name_scope = op.get_attr('ignore_name_scope')
+    process_set_id = op.get_attr('process_set_id')
     recvsplits = op.outputs[1]
 
-    grad_wrt_tensor, _ = alltoall(grad_wrt_output, splits=recvsplits, ignore_name_scope=ignore_name_scope)
+    grad_wrt_tensor, _ = alltoall(grad_wrt_output, splits=recvsplits, ignore_name_scope=ignore_name_scope,
+                                  process_set=_temp_process_set_object(process_set_id))
     grad_wrt_splits = None # not differentiable (integer variable)
 
     return [grad_wrt_tensor, grad_wrt_splits]
 
+
+def _reducescatter(tensor, name=None, op=Sum, ignore_name_scope=False,
+                   process_set=global_process_set):
+    """An op which sums an input tensor over all the Horovod processes, then
+    scatters the result across all the Horovod processes.
+
+    The reduction operation is keyed by the name of the op. The tensor type and
+    shape must be the same on all Horovod processes for a given name. The reduction
+    will not start until all processes are ready to send and receive the tensor.
+
+    Returns:
+      A tensor of the same rank and type as `tensor`. The shape is identical to the
+        input shape, except for the first dimension, which will be divided across
+        the different Horovod processes.
+    """
+    if name is None and not _executing_eagerly():
+        name = 'HorovodReducescatter_%s' % _normalize_name(tensor.name)
+    return MPI_LIB.horovod_reducescatter(tensor, name=name, reduce_op=op,
+                                         ignore_name_scope=ignore_name_scope,
+                                         process_set_id=process_set.process_set_id)
+
+
+@ops.RegisterGradient('HorovodReducescatter')
+def _reducescatter_grad(op, grad):
+    """Gradient for reducescatter op.
+
+    Args:
+      op: An operation.
+      grad: `Tensor` gradient with respect to the output of the op.
+
+    Returns:
+      The gradient with respect to the input of the op.
+    """
+    ignore_name_scope = op.get_attr('ignore_name_scope')
+    process_set_id = op.get_attr('process_set_id')
+    reduce_op = op.get_attr('reduce_op')
+    process_set = _temp_process_set_object(process_set_id)
+    if reduce_op == Sum:
+        grad *= process_set.size()
+    return allgather(grad, ignore_name_scope=ignore_name_scope,
+                     process_set=process_set)
+
+
 def join():
+    """An op to indicate that the rank finished processing data.
+
+    All ranks that did not call join() continue to process allreduce
+    operations. This op is not done before all ranks have joined.
+
+    Returns:
+      An integer scalar containing the the last rank that joined.
+    """
     return MPI_LIB.horovod_join()
 
 
-def size_op(name=None):
+def size_op(process_set_id=0, name=None):
     """An op that returns the number of Horovod processes.
 
     This operation determines the return value at the graph execution time,
@@ -345,12 +460,31 @@ def size_op(name=None):
     Returns:
       An integer scalar containing the number of Horovod processes.
     """
-    return MPI_LIB.horovod_size(name=name)
+    return MPI_LIB.horovod_size(process_set_id=process_set_id, name=name)
 
 
 ops.NotDifferentiable('HorovodSize')
 
 
+def process_set_included_op(process_set_id=0, name=None):
+    """An op that 0 or 1 depending on whether the current process is
+    included in the specified process set or an error code:
+    HOROVOD_PROCESS_SET_ERROR_INIT if Horovod is not initialized,
+    HOROVOD_PROCESS_SET_ERROR_UNKNOWN_SET if the process set is unknown.
+
+    This operation determines the return value at the graph execution time,
+    rather than at the graph construction time, and so allows for a graph to be
+    constructed in a different environment than where it will be executed.
+
+    Returns:
+      An integer scalar with value 0, 1, or an error code.
+    """
+    return MPI_LIB.horovod_process_set_included(process_set_id=process_set_id, name=name)
+
+
+ops.NotDifferentiable('HorovodProcessSetIncluded')
+
+
 def local_size_op(name=None):
     """An op that returns the number of Horovod processes within the
     node the current process is running on.
